@@ -9,15 +9,22 @@ import shutil
 import sys
 import textwrap
 
-from bullet import Bullet
-from bullet import Input
-from bullet import colors
+# TODO(dittrich): https://github.com/Mckinsey666/bullet/issues/2
+# Workaround until bullet has Windows missing 'termios' fix.
+try:
+    from bullet import Bullet
+    from bullet import Input
+    from bullet import colors
+except ModuleNotFoundError:
+    pass
+
 from cliff.command import Command
 from cliff.lister import Lister
 from stat import S_IMODE
 from sys import stdin
 
 
+# TODO(dittrich): Improve this?
 def _is_default(a, b):
     """
     Return "Yes" or "No" depending on whether e is the default
@@ -26,8 +33,74 @@ def _is_default(a, b):
     return "Yes" if a == b else "No"
 
 
+def get_local_default_file():
+    """Returns the path to the local identifier file."""
+    # TODO(dittrich): May need to do this differently to support
+    # Windows file systems.
+    return os.path.join(os.getcwd(), '.python_secrets_environment')
+
+
+def save_default_environment(environment=None):
+    """Save environment identifier to local file for defaulting."""
+    env_file = get_local_default_file()
+    with open(env_file, 'w') as f_out:
+        f_out.write('{0}\n'.format(str(environment)))
+    return True
+
+
+def clear_saved_default_environment():
+    """Remove saved default environment file."""
+    env_file = os.path.join(
+        os.getcwd(),
+        '.python_secrets_environment')
+    if os.path.exists(env_file):
+        os.remove(env_file)
+        return True
+    else:
+        return False
+
+
+def get_saved_default_environment():
+    """Return environment ID value saved in local file or None."""
+    env_file = get_local_default_file()
+    saved_default = None
+    if os.path.exists(env_file):
+        with open(env_file, 'r') as f:
+            saved_default = f.read().replace('\n', '')
+    return saved_default
+
+
+def default_environment():
+    """
+    Return the default environment identifier.
+
+    There are multiple ways for a user to specify the environment
+    to use for python_secrets commands. Some of these involve
+    explicit settings (e.g., via command line option, a
+    saved value in the current working directory, or an
+    environment variable) or implicitly from the name of the
+    current working directory.
+    """
+
+    #  NOTE(dittrich): I know this code has multiple return points
+    #  but it is simpler and easier to understand this way.
+    #
+    # Highest priority is inhereted environment variable.
+    environment = os.getenv('D2_ENVIRONMENT', None)
+    if environment is not None:
+        return environment
+    #
+    # Next is saved file in current working directory.
+    local_default = get_saved_default_environment()
+    if local_default not in ['', None]:
+        return local_default
+    #
+    # Lowest priority is the directory path basename.
+    return os.path.basename(os.getcwd())
+
+
 class EnvironmentsList(Lister):
-    """List the current environments"""
+    """List the current environments."""
 
     LOG = logging.getLogger(__name__)
 
@@ -81,9 +154,10 @@ class EnvironmentsList(Lister):
 
     def take_action(self, parsed_args):
         self.LOG.debug('listing environment(s)')
-        default_environment = psec.secrets.SecretsEnvironment().environment()
+        secrets_environment = psec.secrets.SecretsEnvironment()
+        default_env = default_environment()
         columns = (['Environment', 'Default'])
-        basedir = self.app.secrets.secrets_basedir()
+        basedir = secrets_environment.secrets_basedir()
         if parsed_args.aliasing:
             columns.append('AliasFor')
         data = list()
@@ -92,7 +166,7 @@ class EnvironmentsList(Lister):
             env_path = os.path.join(basedir, e)
             if psec.secrets.is_valid_environment(env_path,
                                                  self.app_args.verbose_level):
-                default = _is_default(e, default_environment)
+                default = _is_default(e, default_env)
                 if not parsed_args.aliasing:
                     item = (e, default)
                 else:
@@ -106,7 +180,7 @@ class EnvironmentsList(Lister):
 
 
 class EnvironmentsCreate(Command):
-    """Create environment(s)"""
+    """Create environment(s)."""
 
     LOG = logging.getLogger(__name__)
 
@@ -220,7 +294,7 @@ class EnvironmentsCreate(Command):
 
 
 class EnvironmentsDelete(Command):
-    """Delete environment"""
+    """Delete environment."""
 
     LOG = logging.getLogger(__name__)
 
@@ -239,7 +313,9 @@ class EnvironmentsDelete(Command):
                             nargs='?',
                             default=None)
         parser.epilog = textwrap.dedent("""
-            Deleting an environment requires use of the ``--force`` flag.
+            Deleting an environment requires use of the ``--force`` flag. If
+            not specified, you will be prompted to confirm the environment
+            name before it is deleted.
 
             .. code-block:: console
 
@@ -274,7 +350,7 @@ class EnvironmentsDelete(Command):
         choice = None
         if parsed_args.environment is not None:
             choice = parsed_args.environment
-        elif not stdin.isatty():
+        elif not (stdin.isatty() and 'Bullet' in globals()):
             # Can't involve user in getting a choice.
             raise RuntimeError('[-] no environment specified to delete')
         else:
@@ -323,7 +399,7 @@ class EnvironmentsDelete(Command):
 
 
 class EnvironmentsRename(Command):
-    """Rename environment"""
+    """Rename environment."""
 
     LOG = logging.getLogger(__name__)
 
@@ -392,7 +468,7 @@ class EnvironmentsRename(Command):
 
 
 class EnvironmentsDefault(Command):
-    """Manage default environment via file in cwd"""
+    """Manage default environment via file in cwd."""
 
     LOG = logging.getLogger(__name__)
 
@@ -470,19 +546,25 @@ class EnvironmentsDefault(Command):
 
     def take_action(self, parsed_args):
         self.LOG.debug('managing localized environment default')
-        cwd = os.getcwd()
-        env_file = os.path.join(cwd, '.python_secrets_environment')
         if parsed_args.unset:
-            try:
-                os.remove(env_file)
-            except Exception as e:  # noqa
-                self.LOG.info('no default environment was set')
+            if parsed_args.environment is not None:
+                raise RuntimeError('--unset does not take an argument')
+            if clear_saved_default_environment():
+                self.LOG.info('explicit default environment unset')
             else:
-                self.LOG.info('default environment unset')
+                self.LOG.info('no default environment was set')
         elif parsed_args.set:
-            if parsed_args.environment is None and not stdin.isatty():
+            # If it is not possible to interactively ask for environment,
+            # just raise an exception.
+            if (
+                parsed_args.environment is None and not
+                    (stdin.isatty() and 'Bullet' in globals())
+            ):
                 raise RuntimeError('[-] no environment specified')
-            if parsed_args.environment is None:
+            # Otherwise, let's prompt for an environment for better UX!
+            if parsed_args.environment is not None:
+                choice = parsed_args.environment
+            else:
                 environments = os.listdir(self.app.secrets.secrets_basedir())
                 choices = ['<CANCEL>'] + sorted(environments)
                 cli = Bullet(prompt="\nChose a new default environment:",
@@ -494,41 +576,46 @@ class EnvironmentsDefault(Command):
                              bullet="→",
                              pad_right=5)
                 choice = cli.launch()
+                # Having second thoughts, eh?
                 if choice == "<CANCEL>":
                     self.LOG.info('cancelled setting default')
-                else:
-                    with open(env_file, 'w') as f:
-                        f.write(choice)
-                    self.LOG.info('default environment set explicitly to ' +
-                                  '"{}"'.format(choice))
-        elif parsed_args.environment is None:
-            # No environment specified, show current setting
-            if os.path.exists(env_file):
-                with open(env_file, 'r') as f:
-                    env_string = f.read().replace('\n', '')
+            if save_default_environment(choice):
+                self.LOG.info(('default environment set explicitly to '
+                               f'"{choice}"'))
+        elif parsed_args.environment is not None:
+            print(parsed_args.environment)
+        else:
+            # No environment specified; show current setting.
+            env_string = get_saved_default_environment()
+            if env_string is not None:
                 if self.app_args.verbose_level > 1:
-                    self.LOG.info('default environment set explicitly to ' +
-                                  '"{}"'.format(env_string))
-                else:
-                    print(env_string)
+                    self.LOG.info(('default environment set explicitly to '
+                                   f'"{env_string}"'))
             else:
+                # No explicit saved default.
+                env_string = default_environment()
                 if self.app_args.verbose_level > 1:
-                    env_string = \
-                            psec.secrets.SecretsEnvironment().environment()
-                    self.LOG.info('default environment is implicitly ' +
-                                  '"{}"'.format(env_string))
-                else:
-                    print(psec.secrets.SecretsEnvironment().environment())
+                    self.LOG.info(('default environment is implicitly '
+                                   f'"{env_string}"'))
+            print(env_string)
 
 
 class EnvironmentsPath(Command):
-    """Return path to files and directories for environment"""
+    """Return path to files and directories for environment."""
 
     LOG = logging.getLogger(__name__)
 
     def get_parser(self, prog_name):
         parser = super().get_parser(prog_name)
         parser.formatter_class = argparse.RawDescriptionHelpFormatter
+        parser.add_argument(
+            '--create',
+            action='store_true',
+            dest='create',
+            default=False,
+            help=("Create the directory path if it does not yet exist "
+                  "(default: False)")
+        )
         parser.add_argument(
             '--exists',
             action='store_true',
@@ -553,12 +640,12 @@ class EnvironmentsPath(Command):
             help='Create and/or return tmpdir for this environment ' +
                  '(default: False)'
         )
-        parser.add_argument('environment',
-                            nargs='?',
+        parser.add_argument('subdir',
+                            nargs='*',
                             default=None)
         parser.epilog = textwrap.dedent("""
             Provides the full absolute path to the environment directory
-            for the environment.
+            for the environment and any specified subdirectories.
 
             .. code-block:: console
 
@@ -569,9 +656,26 @@ class EnvironmentsPath(Command):
 
             ..
 
+            Using the ``--exists`` option will just return ``0`` if the path
+            exists, or ``1`` if it does not. No path is printed on stdout.
+
             Using the ``--tmpdir`` option will return the path to the
             temporary directory for the environment. If it does not already
             exist, it will be created so it is ready for use.
+
+            To append subdirectory components, provide them as arguments and
+            they will be concatenated with the appropriate OS path separator.
+
+            .. code-block:: console
+
+                $ psec environments path -e goSecure configs
+                /Users/dittrich/.secrets/goSecure/configs
+
+            ..
+
+            To ensure the directory path specified by command line arguments
+            is present, use the ``--create`` option.
+
             """)
         return parser
 
@@ -586,18 +690,8 @@ class EnvironmentsPath(Command):
 
     def take_action(self, parsed_args):
         self.LOG.debug('returning environment path')
-        environment = parsed_args.environment
-        if environment is None:
-            environment = self.app.options.environment
+        environment = self.app.options.environment
         e = psec.secrets.SecretsEnvironment(environment)
-        exists = e.environment_exists()
-        if parsed_args.exists:
-            if self.app_args.verbose_level > 1:
-                status = "exists" if exists else "does not exist"
-                self.LOG.info('environment ' +
-                              '"{}" '.format(environment) +
-                              '{}'.format(status))
-            return 0 if exists else 1
         if parsed_args.tmpdir:
             tmpdir = e.tmpdir_path()
             tmpdir_mode = 0o700
@@ -614,18 +708,37 @@ class EnvironmentsPath(Command):
             finally:
                 self._print(tmpdir, parsed_args.json)
         else:
-            self._print(e.environment_path(), parsed_args.json)
+            base_path = e.environment_path()
+            subdir = parsed_args.subdir
+            full_path = base_path if subdir is None \
+                else os.path.join(base_path, *subdir)
+            if not os.path.exists(full_path) and parsed_args.create:
+                mode = 0o700
+                os.makedirs(full_path, mode)
+                if self.app_args.verbose_level > 1:
+                    self.LOG.info(f"created {full_path}")
+            if parsed_args.exists:
+                # Just check existance and return result
+                exists = os.path.exists(full_path)
+                if self.app_args.verbose_level > 1:
+                    status = "exists" if exists else "does not exist"
+                    self.LOG.info('environment path' +
+                                  '"{}" '.format(full_path) +
+                                  '{}'.format(status)
+                                  )
+                return 0 if exists else 1
+            else:
+                self._print(full_path, parsed_args.json)
 
 
 class EnvironmentsTree(Command):
-    """Output tree listing of files/directories in environment"""
+    """Output tree listing of files/directories in environment."""
 
     LOG = logging.getLogger(__name__)
 
     def get_parser(self, prog_name):
         parser = super().get_parser(prog_name)
         parser.formatter_class = argparse.RawDescriptionHelpFormatter
-        default_environment = psec.secrets.SecretsEnvironment().environment()
         parser.add_argument(
             '--no-files',
             action='store_true',
@@ -636,7 +749,7 @@ class EnvironmentsTree(Command):
         )
         parser.add_argument('environment',
                             nargs='?',
-                            default=default_environment)
+                            default=None)
         parser.epilog = textwrap.dedent("""
             The ``environments tree`` command produces output similar
             to the Unix ``tree`` command:
@@ -704,8 +817,11 @@ class EnvironmentsTree(Command):
 
     def take_action(self, parsed_args):
         self.LOG.debug('outputting environment tree')
+        environment = parsed_args.environment
+        if environment is None:
+            environment = self.app.options.environment
         e = psec.secrets.SecretsEnvironment(
-                environment=parsed_args.environment)
+            environment=environment)
         e.requires_environment()
         print_files = bool(parsed_args.no_files is False)
         psec.utils.atree(e.environment_path(),
