@@ -1,9 +1,13 @@
-# Generic modular configuration file manager.
-#
-# Author: Dave Dittrich <dave.dittrich@gmail.com>
-# URL: https://pypi.python.org/pypi/python_secrets
+# -*- coding: utf-8 -*-
 
-"""Python secrets management app"""
+"""
+Python secrets management app
+
+Generic modular configuration file manager.
+
+Author: Dave Dittrich <dave.dittrich@gmail.com>
+URL: https://pypi.python.org/pypi/python_secrets
+"""
 
 from __future__ import print_function
 
@@ -13,10 +17,14 @@ import logging
 import os
 import sys
 import textwrap
+import time
+import webbrowser
 
 from psec import __version__
 from psec import __release__
+from psec.environments import default_environment
 from psec.secrets import SecretsEnvironment
+from psec.utils import bell
 from psec.utils import Timer
 
 # External dependencies.
@@ -47,8 +55,8 @@ def umask(value):
     ivalue = int(value, base=8)
     if ivalue < 0 or ivalue > MAX_UMASK:
         raise argparse.ArgumentTypeError(
-                "value ({}) must be between 0 and " +
-                "0o777".format(value))
+            f"value ({ value }) must be between 0 and 0o777"
+        )
     return ivalue
 
 
@@ -57,7 +65,7 @@ logger = logging.getLogger(__name__)
 
 
 class PythonSecretsApp(App):
-    """Python secrets application class"""
+    """Python secrets application class."""
 
     def __init__(self):
         super(PythonSecretsApp, self).__init__(
@@ -98,14 +106,16 @@ class PythonSecretsApp(App):
                  "(Env: D2_SECRETS_BASEDIR; default: {})".format(
                      _env.secrets_basedir())
         )
+        default_env = default_environment()
         parser.add_argument(
             '-e', '--environment',
             metavar='<environment>',
             dest='environment',
-            default=_env.environment(),
+            default=default_env,
             help="Deployment environment selector " +
                  "(Env: D2_ENVIRONMENT; default: {})".format(
-                     _env.environment())
+                    default_env
+                    )
         )
         parser.add_argument(
             '-s', '--secrets-file',
@@ -131,6 +141,14 @@ class PythonSecretsApp(App):
             help="Export secrets as environment variables " +
                  "(default: False)"
         )
+        parser.add_argument(
+            '--preserve-existing',
+            action='store_true',
+            dest='preserve_existing',
+            default=False,
+            help=("Don't allow over-writing existing environment variables "
+                  "(default: False)")
+        )
         parser.add_argument(
             '--init',
             action='store_true',
@@ -147,6 +165,14 @@ class PythonSecretsApp(App):
             help="Mask to apply during app execution " +
                  "(default: {:#05o})".format(DEFAULT_UMASK)
         )
+        parser.add_argument(
+            '--rtd',
+            action='store_true',
+            dest='rtd',
+            default=False,
+            help=('Open ReadTheDocs documentation on '
+                  '"help" command (default: False)')
+        )
         parser.epilog = textwrap.dedent("""\
             For programs that inherit values through environment variables, you can
             export secrets using the ``-E`` option to the ``run`` subcommand, e.g.
@@ -158,6 +184,17 @@ class PythonSecretsApp(App):
             directory, this reduces the chance that secrets created during execution
             will end up with overly broad permissions.  If you need to relax these
             permissions, use the ``--umask`` option to apply the desired mask.
+
+            To control the browser that is used with the ``help --rtd`` command,
+            set the BROWSER environment variable (e.g., ``BROWSER=lynx``).
+            See: https://github.com/python/cpython/blob/3.8/Lib/webbrowser.py
+
+            Environment Variables:
+              BROWSER             Default browser for use by webbrowser.open()
+              D2_ENVIRONMENT      Defaults the environment identifier.
+              D2_SECRETS_BASEDIR  Defaults the base directory for storing secrets.
+              D2_SECRETS_BASENAME Defaults the base name for secrets storage files.
+              D2_NO_REDACT        Defaults redaction setting for ``secrets show`` command.
             """.format(DEFAULT_UMASK))  # noqa
 
         return parser
@@ -169,7 +206,31 @@ class PythonSecretsApp(App):
                                'module, which requires Python 3.6 or higher')
 
     def prepare_to_run_command(self, cmd):
-        self.LOG.debug('prepare_to_run_command %s', cmd.__class__.__name__)
+        self.LOG.debug('prepare_to_run_command "{0}"'.format(cmd.cmd_name))
+        #
+        # Process ReadTheDocs web browser request here and then
+        # fall through, which also produces help output on the
+        # command line. The webbrowser module doesn't seem to work
+        # consistently on Ubuntu Linux, so this helps a little
+        # though may be confusing when you don't get a browser opening
+        # up like you expect.
+        rtd_url = 'https://python-secrets.readthedocs.io/en/latest/usage.html'
+        if cmd.cmd_name == 'help' and self.options.rtd:
+            for line in [
+                '[+] Opening online documentation for python_secrets on ReadTheDocs.',   # noqa
+                '[+] If a browser does not open, make sure that you are online and/or',  # noqa
+                '[+] enter the following URL in your chosen browser:',
+                rtd_url,
+            ]:
+                self.LOG.info(line)
+            print('\n\n')
+            bell()
+            time.sleep(3)
+            # TODO(dittrich): Add more specificity
+            # FYI, new= is ignored on Windows per:
+            # https://stackoverflow.com/questions/1997327/python-webbrowser-open-setting-new-0-to-open-in-the-same-browser-window-does  # noqa
+
+            webbrowser.open(rtd_url, new=0, autoraise=True)
         self.timer.start()
         os.umask(self.options.umask)
         self.LOG.debug('using environment "{}"'.format(
@@ -177,7 +238,7 @@ class PythonSecretsApp(App):
         self.environment = self.options.environment
         self.secrets_basedir = self.options.secrets_basedir
         # Don't output error messages when "complete" command used
-        if cmd.__class__.__name__ != 'CompleteCommand':
+        if cmd.cmd_name != 'complete':
             SecretsEnvironment.permissions_check(
                 self.secrets_basedir,
                 verbose_level=self.options.verbose_level,
@@ -188,28 +249,27 @@ class PythonSecretsApp(App):
                 secrets_basedir=self.secrets_basedir,
                 secrets_file=self.secrets_file,
                 export_env_vars=self.options.export_env_vars,
+                preserve_existing=self.options.preserve_existing,
                 verbose_level=self.options.verbose_level,
                 env_var_prefix=self.options.env_var_prefix,
                 )
 
     def clean_up(self, cmd, result, err):
-        self.LOG.debug('clean_up %s', cmd.__class__.__name__)
+        self.LOG.debug('clean_up command "{0}"'.format(cmd.cmd_name))
         if err:
             self.LOG.debug('got an error: %s', err)
-            if self.secrets.changed():
+            if self.secrets is not None and self.secrets.changed():
                 self.LOG.info('not writing secrets out due to error')
-        elif cmd.__class__.__name__ != 'CompleteCommand':
+        elif cmd.cmd_name not in ['help', 'complete']:
             if self.secrets.changed():
                 self.secrets.write_secrets()
             if (self.options.elapsed or
                     (self.options.verbose_level > 1
-                     and cmd.__class__.__name__ != "CompleteCommand")):
+                     and cmd.cmd_name != "complete")):
                 self.timer.stop()
                 elapsed = self.timer.elapsed()
-                self.stdout.write('[+] Elapsed time {}\n'.format(elapsed))
-                if sys.stdout.isatty():
-                    sys.stdout.write('\a')
-                    sys.stdout.flush()
+                self.stderr.write('[+] Elapsed time {}\n'.format(elapsed))
+                bell()
 
 
 def main(argv=sys.argv[1:]):
