@@ -1,5 +1,12 @@
 # -*- coding: utf-8 -*-
 
+"""
+Utility functions.
+
+Author: Dave Dittrich
+URL: https://python_secrets.readthedocs.org.
+"""
+
 import argparse
 import ipaddress
 import json
@@ -9,6 +16,7 @@ import random
 import requests
 import time
 import psec.secrets
+import psutil
 import subprocess  # nosec
 import sys
 import textwrap
@@ -16,6 +24,12 @@ import textwrap
 from anytree import Node
 from anytree import RenderTree
 from bs4 import BeautifulSoup
+# TODO(dittrich): https://github.com/Mckinsey666/bullet/issues/2
+# Workaround until bullet has Windows missing 'termios' fix.
+try:
+    from bullet import Bullet
+except ModuleNotFoundError:
+    pass
 from cliff.command import Command
 from cliff.lister import Lister
 from collections import OrderedDict
@@ -32,10 +46,54 @@ AWS_CONFIG_FILE = os.path.join(os.path.expanduser('~'), '.aws', 'credentials')
 LOG = logging.getLogger(__name__)
 
 
+def bell():
+    if sys.stderr.isatty():
+        sys.stderr.write('\a')
+        sys.stderr.flush()
+
+
+# https://stackoverflow.com/questions/7119630/in-python-how-can-i-get-the-file-system-of-a-given-file-path  # NOQA
+def getmount(mypath):
+    """Return the mount point for mypath."""
+    path_ = os.path.realpath(os.path.abspath(mypath))
+    while path_ != os.path.sep:
+        if os.path.ismount(path_):
+            return path_
+        path_ = os.path.abspath(os.path.join(path_, os.pardir))
+    return path_
+
+
+def getmount_fstype(mypath):
+    """Return the file system type for a specific mount path."""
+    mountpoint = getmount(mypath)
+    return get_fs_type(mountpoint)
+
+
+def get_fs_type(mypath):
+    """Return the file system type for mypath."""
+    root_type = ''
+    for part in psutil.disk_partitions():
+        if part.mountpoint == os.path.sep:
+            root_type = part.fstype
+            continue
+        if mypath.startswith(part.mountpoint):
+            return part.fstype
+    return root_type
+
+
 def remove_other_perms(dst):
     """Make all files in path ``dst`` have ``o-rwx`` permissions."""
-    # TODO(dittrich): Test on Windows. Should work on all Linux.
-    get_output(['chmod', '-R', 'o-rwx', dst])
+    # File permissions on Cygwin/Windows filesystems don't work the
+    # same way as Linux. Don't try to change them.
+    # TODO(dittrich): Is there a Better way to handle perms on Windows?
+    fs_type = get_fs_type(dst)
+    if fs_type in ['NTFS', 'FAT', 'FAT32']:
+        msg = ('[-] {0} has file system type "{1}": '
+               'skipping setting permissions').format(
+                   dst, fs_type)
+        LOG.info(msg)
+    else:
+        get_output(['chmod', '-R', 'o-rwx', dst])
 
 
 def get_output(cmd=['echo', 'NO COMMAND SPECIFIED'],
@@ -53,6 +111,8 @@ def get_output(cmd=['echo', 'NO COMMAND SPECIFIED'],
 
 
 def find(lst, key, value):
+    """Find a dictionary entry from a list of dicts where the
+    key identified by 'key' has the desired value 'value'."""
     for i, dic in enumerate(lst):
         if dic[key] == value:
             return i
@@ -70,6 +130,46 @@ def require_options(options, *args):
     return True
 
 
+def prompt_options(options=[],
+                   by_descr=True,
+                   prompt="Select from the following options"):
+    """Prompt the user for a string using an options array."""
+    if 'Bullet' not in globals():
+        raise RuntimeError('[-] can\'t use Bullet on Windows')
+    try:
+        if type(options[0]) is not dict:
+            raise RuntimeError('options is not a list of dictionaries')
+    except Exception as exc:
+        print(str(exc))
+    choices = ['<CANCEL>'] + [
+                                opt['descr']
+                                if by_descr
+                                else opt['ident']
+                                for opt in options
+                             ]
+    cli = Bullet(prompt='\n{0}'.format(prompt),
+                 choices=choices,
+                 indent=0,
+                 align=2,
+                 margin=1,
+                 shift=0,
+                 bullet="→",
+                 pad_right=5)
+    choice = cli.launch()
+    if choice == "<CANCEL>":
+        LOG.info('cancelled selection of choice')
+        return None
+    selected = psec.utils.find(options,
+                               'descr' if by_descr else 'ident',
+                               choice)
+    # options[selected]
+    # {'descr': 'DigitalOcean', 'ident': 'digitalocean'}
+    try:
+        return options[selected]['ident']
+    except Exception as exc:  # NOQA
+        return None
+
+
 def prompt_string(prompt="Enter a value",
                   default=None):
     """Prompt the user for a string and return it"""
@@ -430,7 +530,7 @@ def atree(dir,
 
     output = []
     for pre, fill, node in RenderTree(root_node):
-        output.append(('{}{}'.format(pre, node.name, file=outfile)))
+        output.append((f'{ pre }{ node.name }'))
     if outfile is not None:
         for line in output:
             print(line, file=outfile)
